@@ -1,29 +1,30 @@
 #!/usr/bin/env python3
-################################################################################
-# Project:  ProjPicker (Projection Picker)
-#           <https://github.com/HuidaeCho/projpicker>
-# Purpose:  This Python script provides the CLI and API for ProjPicker.
-# Authors:  Huidae Cho, Owen Smith
-#           Institute for Environmental and Spatial Analysis
-#           University of North Georgia
-# Since:    May 27, 2021
-#
-# Copyright (C) 2021 Huidae Cho <https://faculty.ung.edu/hcho/> and
-#                    Owen Smith <https://www.gaderian.io/>
-#
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <https://www.gnu.org/licenses/>.
-################################################################################
+"""
+Project:  ProjPicker (Projection Picker)
+          <https://github.com/HuidaeCho/projpicker>
+Authors:  Huidae Cho, Owen Smith
+          Institute for Environmental and Spatial Analysis
+          University of North Georgia
+Since:    May 27, 2021
+
+Copyright (C) 2021 Huidae Cho <https://faculty.ung.edu/hcho/> and
+                   Owen Smith <https://www.gaderian.io/>
+
+This program is free software: you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation, either version 3 of the License, or
+(at your option) any later version.
+
+This program is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this program.  If not, see <https://www.gnu.org/licenses/>.
+
+This module implements the CLI and API of ProjPicker.
+"""
 
 import argparse
 import os
@@ -39,10 +40,12 @@ if __package__ is None or __package__ == "":
     from common import bbox_schema, bbox_columns, get_float, BBox
     import coor_latlon
     import coor_xy
+    import gui
 else:
     from .common import bbox_schema, bbox_columns, get_float, BBox
     from . import coor_latlon
     from . import coor_xy
+    from . import gui
 
 # module path
 module_path = os.path.dirname(__file__)
@@ -914,14 +917,32 @@ def parse_mixed_geoms(geoms):
 ################################################################################
 # queries
 
+def sort_bbox(bbox):
+    """
+    Sort a list of BBox instances by area_sqkm in place after deduplicating
+    data by crs_auth_name and crs_code.
+
+    Args:
+        bbox (list): List of BBox instances.
+    """
+    bbox.sort(key=lambda x: x.crs_auth_name+":"+x.crs_code)
+    for i in reversed(range(len(bbox))):
+        if (i > 0 and
+            bbox[i].crs_auth_name == bbox[i-1].crs_auth_name and
+            bbox[i].crs_code == bbox[i-1].crs_code):
+            del bbox[i]
+    bbox.sort(key=lambda x: x.area_sqkm)
+
+
 def query_point(
         point,
         projpicker_db=None):
     """
     Return a list of BBox instances that completely contain an input point
     geometry. Each BBox instance is a named tuple with all the columns from the
-    bbox table in projpicker.db. If projpicker_db is None (default),
-    get_projpicker_db() is used.
+    bbox table in projpicker.db. Results are sorted by area from the smallest
+    to largest. If projpicker_db is None (default), get_projpicker_db() is
+    used.
 
     Args:
         point (list or str): List of two floats or a parseable point geometry.
@@ -929,7 +950,7 @@ def query_point(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     projpicker_db = get_projpicker_db(projpicker_db)
 
@@ -954,7 +975,7 @@ def query_point_using_bbox(
             See parse_point().
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     point = parse_point(point)
 
@@ -974,8 +995,9 @@ def query_points(
     Return a list of BBox instances that completely contain input point
     geometries. Each BBox instance is a named tuple with all the columns from
     the bbox table in projpicker.db. The "and" query mode performs the
-    intersection of bbox rows while the "or" mode the union. If projpicker_db
-    is None (default), get_projpicker_db() is used.
+    intersection of bbox rows while the "or" mode the union. Results are sorted
+    by area from the smallest to largest. If projpicker_db is None (default),
+    get_projpicker_db() is used.
 
     Args:
         points (list): List of parseable point geometries. See parse_points().
@@ -983,7 +1005,7 @@ def query_points(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If query_mode is not one of "and" or "or".
@@ -997,15 +1019,24 @@ def query_points(
     outbbox = []
 
     first = True
+    sort = False
+
     with sqlite3.connect(projpicker_db) as projpicker_con:
         projpicker_cur = projpicker_con.cursor()
         for point in points:
             if query_mode == "or" or first:
-                outbbox.extend(query_point_using_cursor(projpicker_cur, point))
+                obbox = query_point_using_cursor(projpicker_cur, point)
+                if len(obbox) > 0:
+                    if query_mode == "or" and not sort and len(outbbox) > 0:
+                        sort = True
+                    outbbox.extend(obbox)
                 first = False
             else:
                 outbbox = query_point_using_bbox(outbbox, point)
 
+    if sort:
+        sort_bbox(outbbox)
+
     return outbbox
 
 
@@ -1025,7 +1056,7 @@ def query_points_using_bbox(
         query_mode (str): Query mode (and, or). Defaults to "and".
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If query_mode is not one of "and" or "or".
@@ -1056,15 +1087,16 @@ def query_poly(
     """
     Return a list of BBox instances that completely contain an input poly
     geometry. Each BBox instance is a named tuple with all the columns from the
-    bbox table in projpicker.db. If projpicker_db is None (default),
-    get_projpicker_db() is used.
+    bbox table in projpicker.db. Results are sorted by area from the smallest
+    to largest. If projpicker_db is None (default), get_projpicker_db() is
+    used.
 
     Args:
         poly (list): List of parseable point geometries. See parse_poly().
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     return query_polys([poly], "and", projpicker_db)
 
@@ -1083,7 +1115,7 @@ def query_poly_using_bbox(
         poly (list): List of parseable point geometries. See parse_poly().
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     return query_polys_using_bbox(prevbbox, [poly])
 
@@ -1096,8 +1128,9 @@ def query_polys(
     Return a list of BBox instances that completely contain input poly
     geometries. Each BBox instance is a named tuple with all the columns from
     the bbox table in projpicker.db. The "and" query mode performs the
-    intersection of bbox rows while the "or" mode the union. If projpicker_db
-    is None (default), get_projpicker_db() is used.
+    intersection of bbox rows while the "or" mode the union. Results are sorted
+    by area from the smallest to largest. If projpicker_db is None (default),
+    get_projpicker_db() is used.
 
     Args:
         polys (list): List of parseable poly geometries. See parse_polys().
@@ -1105,7 +1138,7 @@ def query_polys(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     polys = parse_polys(polys)
 
@@ -1129,7 +1162,7 @@ def query_polys_using_bbox(
         query_mode (str): Query mode (and, or). Defaults to "and".
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     polys = parse_polys(polys)
 
@@ -1143,8 +1176,9 @@ def query_bbox(
     """
     Return a list of BBox instances that completely contain an input bbox
     geometry. Each BBox instance is a named tuple with all the columns from the
-    bbox table in projpicker.db. If projpicker_db is None (default),
-    get_projpicker_db() is used.
+    bbox table in projpicker.db. Results are sorted by area from the smallest
+    to largest. If projpicker_db is None (default), get_projpicker_db() is
+    used.
 
     Args:
         bbox (list or str): List of four floats or a parseable str of a bbox
@@ -1152,7 +1186,7 @@ def query_bbox(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     projpicker_db = get_projpicker_db()
 
@@ -1179,7 +1213,7 @@ def query_bbox_using_bbox(
             See parse_bbox().
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     bbox = parse_bbox(bbox)
 
@@ -1199,8 +1233,9 @@ def query_bboxes(
     Return a list of BBox instances that completely contain input bbox
     geometries. Each BBox instance is a named tuple with all the columns from
     the bbox table in projpicker.db. The "and" query mode performs the
-    intersection of bbox rows while the "or" mode the union. If projpicker_db
-    is None (default), get_projpicker_db() is used.
+    intersection of bbox rows while the "or" mode the union. Results are sorted
+    by area from the smallest to largest. If projpicker_db is None (default),
+    get_projpicker_db() is used.
 
     Args:
         bboxes (list): List of parseable bbox geometries. See parse_bboxes().
@@ -1208,7 +1243,7 @@ def query_bboxes(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If query_mode is not one of "and" or "or".
@@ -1226,10 +1261,18 @@ def query_bboxes(
         projpicker_cur = projpicker_con.cursor()
         for bbox in bboxes:
             if query_mode == "or" or first:
-                outbbox.extend(query_bbox_using_cursor(projpicker_cur, bbox))
+                obbox = query_bbox_using_cursor(projpicker_cur, bbox)
+                if len(obbox) > 0:
+                    if query_mode == "or" and not sort and len(outbbox) > 0:
+                        sort = True
+                    outbbox.extend(obbox)
                 first = False
             else:
                 outbbox = query_bbox_using_bbox(outbbox, bbox)
+
+    if sort:
+        sort_bbox(outbbox)
+
     return outbbox
 
 
@@ -1249,7 +1292,7 @@ def query_bboxes_using_bbox(
         query_mode (str): Query mode (and, or). Defaults to "and".
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If query_mode is not one of "and" or "or".
@@ -1282,8 +1325,9 @@ def query_geom(
     Return a list of BBox instances that completely contain an input geometry.
     Each BBox instance is a named tuple with all the columns from the bbox
     table in projpicker.db. The "and" query mode performs the intersection of
-    bbox rows while the "or" mode the union. If projpicker_db is None
-    (default), get_projpicker_db() is used.
+    bbox rows while the "or" mode the union. Results are sorted by area from
+    the smallest to largest. If projpicker_db is None (default),
+    get_projpicker_db() is used.
 
     Args:
         geom (list or str): List or str of a parseable geometry. See
@@ -1293,7 +1337,7 @@ def query_geom(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If geom_type is not one of "point", "poly", or "bbox", or
@@ -1329,7 +1373,7 @@ def query_geom_using_bbox(
             "point".
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If geom_type is not one of "point", "poly", or "bbox", or
@@ -1357,8 +1401,9 @@ def query_geoms(
     Return a list of BBox instances that completely contain input geometries.
     Each BBox instance is a named tuple with all the columns from the bbox
     table in projpicker.db. The "and" query mode performs the intersection of
-    bbox rows while the "or" mode the union. If projpicker_db is None
-    (default), get_projpicker_db() is used.
+    bbox rows while the "or" mode the union. Results are sorted by area from
+    the smallest to largest. If projpicker_db is None (default),
+    get_projpicker_db() is used.
 
     Args:
         geoms (list): List of parseable geometries. See parse_points(),
@@ -1369,7 +1414,7 @@ def query_geoms(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If geom_type is not one of "point", "poly", or "bbox", or
@@ -1410,7 +1455,7 @@ def query_geoms_using_bbox(
         query_mode (str): Query mode (and, or). Defaults to "and".
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
 
     Raises:
         Exception: If geom_type is not one of "point", "poly", or "bbox", or
@@ -1443,8 +1488,9 @@ def query_mixed_geoms(
     and "xy" start the latitude-longitude and x-y coordinate systems,
     respectively. This function ignores the current coordinate system set by
     set_coordinate_system(), set_latlon(), or set_xy(), and always starts in
-    the latitude-longitude coordinate system by default. If projpicker_db is
-    None (default), get_projpicker_db() is used.
+    the latitude-longitude coordinate system by default. Results are sorted by
+    area from the smallest to largest. If projpicker_db is None (default),
+    get_projpicker_db() is used.
 
     Args:
         geoms (list or str): List of "point", "poly", "bbox", "latlon", "xy",
@@ -1454,7 +1500,7 @@ def query_mixed_geoms(
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of queried BBox instances sorted by area.
     """
     geoms = parse_mixed_geoms(geoms)
 
@@ -1478,6 +1524,8 @@ def query_mixed_geoms(
         set_latlon()
 
         first = True
+        sort = False
+
         for geom in geoms:
             if geom in ("point", "poly", "bbox"):
                 geom_type = geom
@@ -1486,7 +1534,11 @@ def query_mixed_geoms(
             elif geom == "xy":
                 set_xy()
             elif query_mode == "or" or first:
-                outbbox.extend(query_geom(geom, geom_type, projpicker_db))
+                obbox = query_geom(geom, geom_type, projpicker_db)
+                if len(obbox) > 0:
+                    if query_mode == "or" and not sort and len(outbbox) > 0:
+                        sort = True
+                    outbbox.extend(obbox)
                 first = False
             else:
                 outbbox = query_geom_using_bbox(outbbox, geom, geom_type)
@@ -1496,20 +1548,25 @@ def query_mixed_geoms(
         elif not was_latlon and is_latlon():
             set_xy()
 
+    if sort:
+        sort_bbox(outbbox)
+
     return outbbox
 
 
 def query_all(projpicker_db=None):
     """
     Return a list of all BBox instances. Each BBox instance is a named tuple
-    with all the columns from the bbox table in projpicker.db. If projpicker_db
+    with all the columns from the bbox table in projpicker.db. Results are not
+    sorted by area because no specific geometries are queried. If projpicker_db
     is None (default), get_projpicker_db() is used.
 
     Args:
         projpicker_db (str): projpicker.db path. Defaults to None.
 
     Returns:
-        list: List of queried BBox instances.
+        list: List of all BBox instances in an unsorted manner straight from
+        the projpicker.db bbox table.
     """
     projpicker_db = get_projpicker_db(projpicker_db)
 
@@ -1613,6 +1670,7 @@ def projpicker(
         print_geoms=False,
         overwrite=False,
         append=False,
+        start_gui=False,
         projpicker_db=None,
         proj_db=None,
         create=False):
@@ -1634,7 +1692,9 @@ def projpicker(
     and separator options only apply to the plain output format. The overwrite
     option applies to both projpicker.db and the output file, but the append
     option only appends to the output file. Only one of the overwrite or append
-    options must be given. If projpicker_db or proj_db is None (default),
+    options must be given. For selecting a subset of queried BBox instances, a
+    GUI can be launched by setting gui to True. Results are sorted by area from
+    the smallest to largest. If projpicker_db or proj_db is None (default),
     get_projpicker_db() or get_proj_db() is used, respectively.
 
     Args:
@@ -1651,6 +1711,8 @@ def projpicker(
             False.
         append (bool): Whether or not to append output to file. Defaults to
             False.
+        gui (bool): Whether or not to start a GUI for selecting part of queries
+            BBox instances. Defaults to False.
         projpicker_db (str): projpicker.db path. Defaults to None.
         proj_db (str): proj.db path. Defaults to None.
         create (bool): Whether or not to create a new projpicker.db. Defaults
@@ -1695,6 +1757,9 @@ def projpicker(
 
     bbox = query_mixed_geoms(geoms, projpicker_db)
 
+    if start_gui:
+        bbox = gui.select_bbox(bbox)
+
     mode = "w"
     header = not no_header
     if append and outfile != "-" and os.path.exists(outfile):
@@ -1810,6 +1875,10 @@ def parse():
             "-o", "--output",
             default="-",
             help="output bbox file path (default: stdout); use - for stdout")
+    parser.add_argument(
+            "-g", "--gui",
+            action="store_true",
+            help="start GUI for selecting CRSs")
     parser.add_argument(
             "geometry",
             nargs="*",
@@ -1839,6 +1908,7 @@ def main():
     separator = args.separator
     infile = args.input
     outfile = args.output
+    start_gui = args.gui
     geoms = args.geometry
 
     if version:
@@ -1860,6 +1930,7 @@ There is NO WARRANTY, to the extent permitted by law.""")
             print_geoms,
             overwrite,
             append,
+            start_gui,
             projpicker_db,
             proj_db,
             create)
